"""
Module providing typed dicts for the metadata HandlerMeta:
- pest/metadata/types/handler_meta.py

ATTENTION:
This file was auto-generated by `task gen:types`.
Do not edit manually.
"""

from enum import Enum
from typing import Any, Dict, List, Sequence, Type, TypedDict, Union

from fastapi import Response


class HandlerMetaDict(TypedDict, total=False):
    response_model: Union[Any, None]
    '''🐀 ⇝ response model of the handler'''

    response_class: Union[Type[Response], None]
    '''🐀 ⇝ response class of the handler'''

    status_code: Union[int, None]
    '''🐀 ⇝ status code of the handler'''

    response_model_exclude_none: Union[bool, None]
    '''🐀 ⇝ exclude none values from the response model'''

    tags: Union[List[Union[str, Enum]], None]
    '''🐀 ⇝ tags of the handler'''

    dependencies: Union[Sequence[Any], None]
    '''🐀 ⇝ dependencies of the handler'''

    summary: Union[str, None]
    '''🐀 ⇝ summary of the handler'' '''

    description: Union[str, None]
    '''🐀 ⇝ description of the handler'''

    deprecated: Union[bool, None]
    '''🐀 ⇝ is the handler deprecated?'''

    name: Union[str, None]
    '''🐀 ⇝ name of the handler'''

    responses: Union[Dict[Union[int, str], Dict[str, Any]], None]
    '''🐀 ⇝ responses of the handler'''
